#!/usr/bin/perl -w

# usenet new gateway hack for phpBB

# Database specs
my $db          = "dirtbike_drupal";
my $db_host     = "localhost";
my $db_username = "";
my $db_password = "";

my $newsserver = { 
    server   => "news.readfreenews.net",
    username => "",
    password => ""
    };

my $maxmsglen = 20000;    # Messages bigger than this many lines get skipped
my $maxuserlength = 100;  # Maximum length of users email address
my $batch_size = 100;     # Number of articles to grab before writing to the db
my $expire_days = 7;      # Days to keep posts before expiring
my $hyperlinks = 1;       # Put <A HREF=></A> tags on hyperlinks
my $custom_quotes = 1;    # Do Quoted indenting
my $rightmargin = 79;     # Right margin for auto text wrapping
my $maxmsgs = 50;       # Maximum messages to retrieve from a newsgroup

my $quote3 = "green";     # Quote level 3
my $quote2 = "darkred";   # Quote level 2
my $quote1 = "blue";      # Quote level 1

# Pid file name - change if you want
my $pid_file = "newsgrab.pid";

# Default newgateway user id, change if you create a specific news user
my $newsgatewayid = "-2";

# Posting Settings
my $include_sig      = 1; # Send users signatures?
my $send_users_email = 0; # Allow sending out of users email addresses
my $posting_enabled  = 1; # This must be on to post out
my $useragent        = "Newsgateway";
my $organisation     = "Your organisation";
my $nntp_from        = "newsgateway-no.spam\@dirtbike.spam.go.away.ws";
my $msgid_domain     = "\@dirtbike.spam.go.away.ws";
my $site_footer      = "Posted via the newsgateway at http://www.dirtbike.ws/";

my $phpbbver1    = 0;
my $phpbbver2    = 1;

# Stuff to change if you use a different phpBB derivative
# Standard phpBB
#my $forums      = "forums";
#my $categories  = "categories";
#my $posts       = "posts";
#my $posts_text  = "posts_text";
#my $topics      = "topics";
#my $users       = "users";

# phpBBPlus/V2
my $forums      = "phpbb_forums";
my $categories  = "phpbb_categories";
my $posts       = "phpbb_posts";
my $posts_text  = "phpbb_posts_text";
my $topics      = "phpbb_topics";
my $users       = "phpbb_users";

# Downloading image support
my $posts_images = "phpbb_posts_images";
my $retrieve_images = 0;
undef $/;                 # Slurp whole files at once

# Testing stuff here
my $testingzap  = 1;      # Zap the database contents each run! Careful!
my $ignorepid   = 0;      # Ignore the status of the pid file
my $debug       = 1;      # Debug level, 0 = errors only, 1 = lots, 2 = heaps

# You should not need to change anything below here.
use strict;
use News::NNTPClient;
use Net::SMTP;
use MIME::WordDecoder;
use Text::Autoformat;
use Mail::Address;
use URI::Find;
use DBI;
use Date::Parse;
use Date::Manip;
use HTML::Entities;
use String::Random;

use File::Temp qw/ tempfile tempdir /;

&Date_Init("TZ=CAST");    # man Date::Manip and set your own here

console("Checking for other running instances of newsgrab\n");
my $pid=$$;
my $old_pid=0;
if (-e $pid_file && !$ignorepid) {
    $| = 1;
    print "Newsgrab is already running!\n";
    print "Or maybe died without cleaning up\n";
    print "Remove $pid_file from the newsgrab directory to allow execution\n";
    die "Exiting\n";
} else {
    open(W,">$pid_file");
    print W $pid;
    close(W);
}

# Determine current directory.
my $currdir;
if ($0=~m#^(.*)(\\|/)#) { 
    $currdir = $1; 
} else { 
    $currdir = `pwd`; 
    chomp $currdir; 
}

console("Initialising Word Decoder\n");
my $wd = supported MIME::WordDecoder "ISO-8859-1";

console("Connecting to database\n");
my $dbh = db_connect();

console("Connecting to newsserver\n");
my $news = connect_usenet(\$newsserver);

# Get the list of newsgroups to read, uses forum_name as the newsgroup name
console("Fetching list of newsgroups\n");
my $qry = "";
if ($phpbbver1) {
    $qry = db_fetch("SELECT forum_id, forum_name, usenet_lastpost " .
		    "FROM $forums " .
		    "WHERE forum_type = 2 ");
} elsif ($phpbbver2) {
    $qry = db_fetch("SELECT a.forum_id, a.forum_name, a.usenet_lastpost " .
		    "FROM $forums as a, $categories as b " .
		    "WHERE (a.cat_id = b.cat_id) " .
		    "AND b.cat_title = 'Newsgroups'");
} else {
    die ("Please configure a phpbbversion in newsgrab.pl - read\n" .
	 "step 4. in phpBB-INSTALL.txt\n");
}

# Process the newsgroups one by one
while (my($forum_id, $forum_name, $usenet_lastpost) = $qry->fetchrow_array()) {
    my ($firstmsg, $lastmsg, $tmp) = $news->group($forum_name);
    
    if ($testingzap) {
	console("Zapping $forum_name - $forum_id\n");
	db_execute("DELETE from $topics " .
		   "WHERE forum_id = $forum_id");
	
	my $qry2 = db_fetch("SELECT post_id from $posts " .
			    "WHERE forum_id = $forum_id");
	while (my($post_id) = $qry2->fetchrow_array()) {
	    db_execute("DELETE from $posts_text " .
		       "WHERE post_id = $post_id");
	}
	
	db_execute("DELETE from $posts " .
		   "WHERE forum_id = $forum_id");
	db_execute("UPDATE $forums " .
		   "SET forum_posts = 0, usenet_lastpost = 0, " .
		   "forum_topics = 0, forum_last_post_id = 0 " .
		   "WHERE forum_id = $forum_id");
	db_execute("UPDATE $users " .
		   "SET user_posts = 0 " .
		   "WHERE user_id = $newsgatewayid");
	$usenet_lastpost = 0;
    }
    
    if (!defined($usenet_lastpost) || $usenet_lastpost == 0) {
	$usenet_lastpost = $firstmsg;
    } elsif ($usenet_lastpost > $lastmsg) {
	$usenet_lastpost = $lastmsg;
    }
    
    console("Grabbing newsgroup: $forum_name...\n");
    
    if ($lastmsg > $usenet_lastpost) {
	if (($lastmsg - $usenet_lastpost) > $maxmsgs) {
	    $usenet_lastpost = $lastmsg - $maxmsgs;
	}
	for (my $i = $usenet_lastpost; $i < $lastmsg; $i += $batch_size) {
	    my @articles = get_batch($i, $i + $batch_size, $news);
	    
	    if ($#articles >= 0) {
		console("Retrieved " . ($#articles + 1) . " articles\n");
		article_batch_process(\@articles, $forum_id, $forum_name);
	    }
	}
	article_forum_load_t($forum_id, $forum_name);
	console(" -> finding replies... \n");
	article_forum_load_p($forum_id, $forum_name);
	db_execute("UPDATE $forums " .
		   "SET usenet_lastpost = $lastmsg " .
		   "WHERE forum_id = '$forum_id'");
    } else {
	console(" No new messages.\n\n");
    }
    post_outgoing_p($news, $forum_id, $forum_name);
}

article_expire();
$dbh->disconnect;
unlink $pid_file;

#
# Functions below
#

sub post_outgoing_p {
    my $c = $_[0];
    my $forum_id = $_[1];
    my $forum_name = $_[2];
    my ($email,$signature,$fld,$sfld);
    my ($inreplys, $replytomsgid, $refs, $subject);

    my $sql_posts = db_fetch("SELECT b.user_email, b.user_viewemail, " .
			     "b.user_sig, a.post_id, a.topic_id, a.inreplyto, " .
			     "b.username, e.forum_name, d.topic_title, " .
			     "d.topic_id, c.post_text " .
			     "FROM $posts AS a, $users AS b, $posts_text as c, " .
			     "$topics AS d, $forums AS e " .
			     "WHERE a.poster_id = b.user_id " .
			     "AND a.topic_id = d.topic_id " .
			     "AND c.post_id = a.post_id " .
			     "AND d.forum_id = e.forum_id " .
			     "AND e.forum_id = $forum_id " .
			     "AND a.isusenetpost = 0");
    db_execute("begin");
    while (my $post = $sql_posts->fetchrow_hashref()) {
	my $poster = $dbh->quote($post->{'username'});
	my $newsgroup = $dbh->quote($post->{'forum_name'});
	if ($post->{'inreplyto'}) {
	    $subject = $dbh->quote("Re: ".$post->{'topic_title'});
	} else {
	    $subject = $dbh->quote($post->{'topic_title'});
	}
	my $body = $post->{'post_text'};
	if ($include_sig) {
	    $signature = $post->{'user_sig'};
	    # If the user actually has one configured
	    if ($signature) {
		$signature = remove_bb_code($signature);
	    } else {
		$signature = "";
	    }
	} else {
	    $signature = "";
	}
	if ($send_users_email) {
	    if ($post->{'user_email'}) {
		$email = $post->{'user_email'};
	    } else {
		$email = $nntp_from;
	    }
	} else {
	    $email = $nntp_from;
	}
	
	$email = $dbh->quote($email);
	$signature = $dbh->quote($signature);
	
	my $threads = db_fetch("SELECT msgid, topic_title " .
			       "FROM $topics " .
			       "WHERE topic_id = $post->{'topic_id'}");
	my $thread = $threads->fetchrow_hashref();

	if ($post->{'inreplyto'}) {
	    $inreplys = db_fetch("SELECT msgid " .
				 "FROM $posts " .
				 "WHERE post_id = $post->{'inreplyto'}");
	    $replytomsgid = $inreplys->fetchrow_array;
	} else {
	    $replytomsgid = "";
	}
	
	if (($thread->{msgid} eq "") 
	    and $replytomsgid eq "") {
	    # Must be a new message from our side, make up a msg id
	    my $foo = new String::Random;
	    $foo->{'A'} = [ 'A'..'Z', 'a'..'z','0'..'9' ];
	    my $msgid = $dbh->quote("<" . 
				    $foo->randpattern("AAAAAAAAAAAAAAAAAA") . 
				    $msgid_domain .
				    ">");
	    # Since we've created a new "msgid", we need to update our
	    # topic now
	    db_execute("UPDATE $topics " .
		       "SET msgid = $msgid " .
		       "WHERE topic_id = $post->{'topic_id'}");
	    db_execute("UPDATE $posts " .
		       "SET msgid = $msgid " .
		       "WHERE topic_id = $post->{'topic_id'} " .
		       "AND post_id = $post->{'post_id'}");
	    $refs = $dbh->quote("$msgid");
	} else {
	    $refs = $dbh->quote("$thread->{msgid} $replytomsgid");
	}
	unless ($body eq "") {
	    $body = $dbh->quote($body);
	    db_execute("INSERT IGNORE INTO phpbb_usenet_outgoing " .
		       "(poster,email,signature,newsgroup,subject," .
		       "refs,body,threadid,postid,inreplyto) " .
		       "VALUES " .
		       "($poster,$email,$signature,$newsgroup,$subject," .
		       "$refs,$body,$post->{'topic_id'}," .
		       "$post->{'post_id'},$post->{'inreplyto'})");
	}
	db_execute("UPDATE $posts SET " . 
		   "isusenetpost = -1, nntpdateline = " . time() . " " .
		   "WHERE post_id = $post->{'post_id'}");
    }
    db_execute("commit");
    $sql_posts->finish;
    post_outgoing($c,$forum_id,$forum_name);
}

sub post_outgoing {
    my $c = $_[0];
    my $forum_id = $_[1];
    my $forum_name = $_[2];
    my @headers=();
    my @article=();
    my (@r,$fbody,$fquote,$qp);
    
    my $q2 = db_fetch("SELECT poster,email,signature,newsgroup," .
		      "subject,refs,body,threadid,postid,inreplyto " .
		      "FROM phpbb_usenet_outgoing " .
		      "WHERE (refs <> ' ' OR subject NOT LIKE 'Re: %') " .
		      "AND newsgroup = ". $dbh->quote($forum_name));
    
    while (my ($poster,$email,$signature,$group,$subject,$refs,
	       $body,$threadid,$postid,$inreplyto) = $q2->fetchrow_array) {
	my $quote = $body;
	$quote =~ m/\[QUOTE\](.*)\[\/QUOTE\]/iso; 
	$quote = $1;
	if ($quote) {
	    my $regex = quotemeta($quote);
	    $body =~ s/$regex//s;
	    $quote =~ m/Originally posted by (.*)\[\/i\]/so; 
	    $qp = $1;
	    $quote =~ s/\[i\]Originally posted by.*\[\/i\]//iso;
	    $quote = remove_bb_code($quote);
	}
	if ($refs) {
	    @r = split(/\s/,$refs);
	}
	my $re = "";
	foreach my $ref (@r) {
	    if ($ref =~ /^</o) {
		$re = $re . " " . $ref;
	    }
	}
	push(@headers,"Newsgroups: $group\n");
	push(@headers,"Subject: $subject\n");
	push(@headers,"From: $poster <$email>\n");
	push(@headers,"Sender: $poster\n");
	push(@headers,"User-Agent: $useragent\n");
	if ($re) {
	    push(@headers,"References: $re\n");
	}
	push(@headers,"Organization: $organisation\n");
	push(@headers,"X-Newsreader: $useragent\n");
	push(@headers,"X-References: $threadid|$postid\n");
	$body = remove_bb_code($body);
	if ($body =~ /\w/o) {
	    $fbody = autoformat $body, {right=>72,all=>1};
	}
	if ($quote) {
	    $fquote = autoformat $quote, {right=>72,all=>1};
	    $fquote =~ s/^\n\s+$|^\n$|^>$|^>\s+$|^\s+>\s+$//gos;
	    my @quotes =  split(/\n/,$fquote);
	    push(@article,"\n");
	    push(@article,"$qp wrote:\n");
	    foreach my $q (@quotes) {
		if ($q) {
		    push(@article,"  > $q\n\n");
		}
	    }
	}
	push(@article,"$fbody\n");
	push(@article,"\n\n--\n" . $site_footer . "\n");
	if ($posting_enabled) {
	    console("Posting message by $poster to $group...\n");
	    if ($poster && $body) {
		if ($debug > 1) {
		    print @headers;
		    print "\n";
		    print @article;
		} else {
		    print @headers;
		    print "\n";
		    print @article;
		    $c->post(@headers, "", @article);
		    console(" Sent\n");
		}
	    }
	}
	db_execute("DELETE FROM phpbb_usenet_outgoing WHERE postid = $postid");
	$poster = $dbh->quote($poster);
	$group = $dbh->quote($group);
	$subject = $dbh->quote($subject);
	$refs = $dbh->quote($refs);
	$body = $dbh->quote($body);
	$signature = $dbh->quote($signature);
	$email = $dbh->quote($email);
	my $dtm = time();
	db_execute("INSERT INTO phpbb_usenet_outgoing_log " .
		   "(dtm,poster,email,signature,newsgroup,subject," .
		   "refs,body,threadid,postid,inreplyto) " .
		   "VALUES " .
		   "($dtm,$poster,$email,$signature,$group,$subject," .
		   "$refs,$body,$threadid,$postid,$inreplyto)");
	@article=(); @r=();
    } 
    $q2->finish;
}

sub article_forum_load_p {
    my $forum_id = $_[0];
    my $forum_name = $_[1];
    my $t = time();
    my $lastpostdtm;

    my $found_image;
    my ($post_image, $post_image_filename);
    my ($tempdir, $tempfile, $fh);

    # check for orphan replies
    db_execute("begin");
    
    my $qo = db_fetch("SELECT msgid, subject " .
		      "FROM phpbb_usenet_article " .
		      "WHERE forum = $forum_id " .
		      "AND subject <> '(no subject)' " .
		      "AND LEFT(subject,3) = 'Re:' " .
		      "AND (refs = '' OR refs is NULL OR refs = '(none)')");
    
    while (my($msgid,$subject) = $qo->fetchrow_array) {
		my $m; 
		my @am;
		$msgid = $dbh->quote($msgid);
		$subject =~ s/^Re:\s+//io;
		my $tq  = db_fetch("SELECT STRAIGHT_JOIN a.msgid " .
				   "FROM $topics AS b, $posts AS a " .
				   "WHERE b.forum_id = $forum_id " .
				   "AND b.topic_id = a.topic_id AND b.topic_title =" . 
				   $dbh->quote($subject));
		$m = "";
		while (my ($tref) = $tq->fetchrow_array) {
			$m .= " $tref";
			$tref = $dbh->quote($tref);
			db_execute("INSERT IGNORE INTO phpbb_usenet_ref " .
				   "(msgid,ref,dtm) " .
				   "VALUES ($msgid,$tref,$t)");
		} 
		$tq->finish;
		if ($m ne "") {
			$m =~ s/^\s+|\s+$//gso;
			@am = split(/ /,$m);
		}
		$m = $dbh->quote($m);
		db_execute("UPDATE phpbb_usenet_article " .
			   "SET refs = $m, ord_key = $#am+1 " .
			   "WHERE msgid = $msgid");
    }
    $qo->finish;
	
    # load replies
    db_execute("TRUNCATE phpbb_usenet_article_loader");

    db_execute("INSERT IGNORE INTO phpbb_usenet_article_loader " .
	       "SELECT DISTINCT a.newsgroup, a.forum, a.msgid, a.dtm, " .
	       "a.subject, a.poster, a.email, a.nntpposter, a.refs, a.body, " .
	       "a.msgnum, a.ord_key, a.threadid, a.postid " .
	       "FROM phpbb_usenet_article AS a " .
	       "left outer join $posts AS b ON (a.msgid = b.msgid) " .
	       "WHERE b.msgid IS NULL AND a.forum = $forum_id");
    
    my $isreplies = 1;
    my $showreplies = 1;
    my @updated_threads=();
	
    while ($isreplies) {
	if ($showreplies == 1) {
	    console(" -> loading replies..."); 
	    $showreplies = 0;
	}
	my $q3 = db_fetch("SELECT count(distinct(a.msgid)) AS cnt " .
			  "FROM phpbb_usenet_article_loader AS a, " .
			  "$topics AS b, phpbb_usenet_ref AS c, $posts AS d " .
			  "where b.topic_id = d.topic_id " .
			  "and c.ref = d.msgid and a.msgid = c.msgid " .
			  "AND b.forum_id = $forum_id AND d.ord_key <= a.ord_key");
	$isreplies = $q3->fetchrow_array; $q3->finish;
	$isreplies && console("      $isreplies... ");
	$q3 = db_fetch("SELECT DISTINCT " .
		       "b.topic_title, a.nntpposter, a.forum, a.msgid, a.dtm, " .
		       "a.subject, a.poster, a.body, a.ord_key, a.postid, a.email, b.topic_id " .
		       "FROM phpbb_usenet_article_loader AS a, " .
		       "$topics AS b, phpbb_usenet_ref AS c, $posts AS d " .
		       "where b.topic_id = d.topic_id " .
		       "and c.ref = d.msgid and a.msgid = c.msgid " .
		       "AND b.forum_id = $forum_id AND d.ord_key <= a.ord_key " .
		       "ORDER BY a.ord_key, a.dtm");
	my $fbody;
	while ( my ($title,$nntpposter,$forumid,$msgid,$nntpdtm,$subject,$poster,$body,$ord,$postid,$email,$threadid) = $q3->fetchrow_array ) {
	    my $littlesubj;
	    if ((($body =~ /begin/oi) ||
		 ($body =~ /--Next_Part--/oi) ||
		 ($body =~ /This is a multi-part message in MIME format/oi))
		&& ($retrieve_images)) {
		console("Found Begin, processing\n");
		$tempdir = tempdir("ngXXXXXX");
		($fh, $tempfile) = tempfile("ngpXXXXXX", 
					    DIR => $tempdir,
					    UNLINK => 0);
		print $fh $body;
		close $fh;
		`/usr/local/bin/uudeview -m -i -o -q -n -t -p $tempdir $tempfile`;
		`/bin/rm $tempfile`;
		
		# Set the body to blank
		$body = "";
		$found_image = 1;
	    }
	    if ($body =~ /\w/o) {
		$body = autoformat $body, {right=>$rightmargin,all=>1};
	    }
	    $body = article_add_html($body);
	    $body = $dbh->quote($body);
	    if(length($title) > 35) {
		$littlesubj = substr($title,0,35)."...";
	    } else {
		$littlesubj = $title;
	    }
	    $littlesubj = $dbh->quote($littlesubj);
	    $subject = $dbh->quote($subject);
	    if (!$poster) {
		$poster = "Anonymous";
	    }
	    $poster = $dbh->quote($poster);
	    if (!$email) {
		$email = $nntp_from;
	    }
	    $email = $dbh->quote($email);
	    $msgid = $dbh->quote($msgid);
	    if (!$nntpposter) {
		$nntpposter = "Unknown";
	    }
	    $nntpposter =~ s/<|>/\#/go; 
	    $nntpposter = $dbh->quote($nntpposter);
	    if ($postid > 0) {
		db_execute("UPDATE $posts " .
			   "SET msgid = $msgid " .
			   "WHERE post_id = $postid");
		$postid = $dbh->{'mysql_insertid'};
	    } else {
		db_execute ("UPDATE phpbb_usenet_outgoing_log " .
			    "SET msgid = $msgid " .
			    "WHERE postid = $postid");
		db_execute ("INSERT INTO $posts " .
			    "(topic_id,forum_id,post_time,nntpdateline,msgid,ord_key,poster_id," .
			    "nntpfrom,poster_ip,isusenetpost) " .
			    "VALUES " .
			    "($threadid,$forumid,'" . 
			    &UnixDate("today", "%Y-%m-%d %H:%M") . 
			    "',$nntpdtm,$msgid,'0',$newsgatewayid,$nntpposter,0,1)");
		$postid = $dbh->{'mysql_insertid'};
		db_execute ("UPDATE $forums " .
			    "SET forum_last_post_id = $postid " .
			    "WHERE forum_id = $forumid");
	    }
	    
	    if ($found_image) {
		opendir(DIR, $tempdir);
		my @allfiles = grep !/^\./, readdir(\*DIR);
		foreach my $filename (@allfiles) {
		    console("Found attachment: $filename\n");
		    
		    # Now load the file from disk
		    console("Loading $tempdir/$filename\n");
		    open(INPUT, $tempdir . "/" . $filename);
		    $post_image = readline(\*INPUT);
		    $post_image_filename = $filename;
		    close(INPUT);
		    `/bin/rm $tempdir/"$filename"`;
					
		    if (($filename =~ /jpg/i)
			|| ($filename =~ /gif/i)
			|| ($filename =~ /png/i)) {
			
			console("Found image name: $filename\n");
			db_execute("INSERT INTO $posts_images " .
				   "(post_id, post_image, " .
				   "filename, post_image_index) " .
				   "VALUES " .
				   "($postid, " .
				   $dbh->quote($post_image) .
				   ", '$post_image_filename', " .
				   "'$found_image')");
			$found_image++;
		    } elsif ($filename =~ /txt/i) {
			$body = $dbh->quote($post_image);
		    }
		}
		closedir(DIR);
		
		db_execute("INSERT INTO $posts_text " .
			   "(post_id, post_text) " .
			   "VALUES " .
			   "($postid, $body)");
		
		undef $found_image;
		`/bin/rmdir $tempdir`;
	    } else {
		db_execute("INSERT INTO $posts_text " .
			   "(post_id, post_text) " .
			   "VALUES " .
			   "($postid, $body)");
	    }
	    
            db_execute("DELETE FROM phpbb_usenet_article WHERE msgid = $msgid");
            db_execute("DELETE FROM phpbb_usenet_article_loader WHERE msgid = $msgid");
            db_execute("DELETE FROM phpbb_usenet_ref WHERE msgid = $msgid");
	    
	    db_execute("UPDATE $topics " .
		       "SET topic_replies = topic_replies + 1, " .
		       "topic_last_post_id = $postid " .
		       "WHERE topic_id = $threadid");
	    
            db_execute("UPDATE $forums " .
		       "SET forum_posts = forum_posts + 1 " .
		       "WHERE forum_id = $forumid");
	    
            db_execute("UPDATE $users " .
                       "SET user_posts = user_posts + 1 " .
                       "WHERE user_id = $newsgatewayid");
			
            push(@updated_threads,$threadid);
	}
    }
    db_execute("commit");
    console("\n");
    return @updated_threads;
}

sub article_forum_load_t {
    my $forum_id = $_[0];
    my $forum_name = $_[1];
    my ($littlesubj,$lastpostdtm);
    my $tcount = 0;
    my $found_image;
    my ($post_image, $post_image_filename);
    my ($tempdir, $tempfile, $fh);

    # load thread starters
    console(" -> inserting new threads from $forum_name\n");
    my $q1 = db_fetch("SELECT a.nntpposter, a.forum, a.msgid, " .
		      "a.dtm, a.subject, a.poster, a.email, " .
		      "a.body, a.ord_key, a.threadid, a.postid " .
		      "FROM phpbb_usenet_article AS a " .
		      "LEFT OUTER JOIN $posts AS b on (a.msgid=b.msgid) " .
		      "WHERE a.forum = $forum_id " .
		      "AND b.msgid is null " .
		      "AND (refs = '' OR refs is NULL OR refs = '(none)') " .
		      "AND LEFT(subject,3) <> 'Re:' ORDER BY dtm");
    db_execute("begin");
    while ( my ($nntpposter,$forumid,$msgid,$nntpdtm,$subject,
		$poster,$email,$body,$ord,$threadid,$postid) 
	    = $q1->fetchrow_array ) {
	$tcount++;
	
	if ((($body =~ /begin/oi) ||
	     ($body =~ /--Next_Part--/oi) ||
	     ($body =~ /This is a multi-part message in MIME format/oi))
	    && ($retrieve_images)) {
	    console("Found Begin, processing\n");
	    $tempdir = tempdir("ngXXXXXX");
	    ($fh, $tempfile) = tempfile("ngpXXXXXX", 
					DIR => $tempdir,
					UNLINK => 0);
	    print $fh $body;
	    close $fh;
	    `/usr/local/bin/uudeview -m -i -o -q -n -t -p $tempdir $tempfile`;
	    `/bin/rm $tempfile`;
	    
	    # Set the body to blank
	    $body = "";
	    $found_image = 1;
	}
	if ($body =~ /\w/o) {
	    $body = autoformat $body, {right=>$rightmargin,all=>1};
	}
	$body = article_add_html($body);
	$body = $dbh->quote($body);
	if (!$email) {
	    $email = $nntp_from;
	}
	$email = $dbh->quote($email);
	if(length($subject) > 35) {
	    $littlesubj = substr($subject,0,35)."...";
	} else {
	    $littlesubj = $subject;
	}
	$littlesubj = $dbh->quote($littlesubj);
	$subject = $dbh->quote($subject);
	$poster =~ s/^\s+|\s+$//go;
	if ($poster eq "") {
	    $poster = "Anonymous";
	}
	$poster = $dbh->quote($poster);
	$msgid = $dbh->quote($msgid);
	if (!$nntpposter) {
	    $nntpposter = "Unknown";
	}
	$nntpposter =~ s/<|>/\#/go; 
	$nntpposter = $dbh->quote($nntpposter);
	if ($threadid > 0) {
	    db_execute ("UPDATE $topics " .
			"SET msgid = $msgid " .
			"WHERE topic_id = $threadid");
	} else {
	    $lastpostdtm = &UnixDate($nntpdtm, "%Y-%m-%d %H:%M");
	    db_execute ("INSERT INTO $topics " .
			"(topic_title,topic_poster,topic_time," .
			"forum_id,nntpdateline,msgid,isusenetpost,regpost) " .
			"VALUES " .
			"($subject,$newsgatewayid,'" . 
			&UnixDate("today", "%Y-%m-%d %H:%M") .
			"',$forumid,$nntpdtm,$msgid,1,0)");
	    $threadid = $dbh->{'mysql_insertid'};
	    
	    db_execute("INSERT INTO $posts " .
		       "(topic_id,forum_id,post_time,nntpdateline,msgid,ord_key," .
		       "poster_id,nntpfrom,poster_ip,isusenetpost) " .
		       "VALUES " .
		       "($threadid,$forumid,'" . 
		       &UnixDate("today", "%Y-%m-%d %H:%M") . 
		       "',$nntpdtm,$msgid,'0',$newsgatewayid,$nntpposter,0,1)");
	    $postid = $dbh->{'mysql_insertid'};
	    
	    db_execute("UPDATE $topics " .
		       "SET topic_last_post_id = $postid, topic_first_post_id = $postid " .
		       "WHERE topic_id = $threadid");
	    
	    db_execute("UPDATE $forums " .
		       "SET forum_last_post_id = $postid " .
		       "WHERE forum_id = $forumid");
	    
	    if ($found_image) {
		opendir(DIR, $tempdir);
		my @allfiles = grep !/^\./, readdir(\*DIR);
		foreach my $filename (@allfiles) {
		    console("Found attachment: $filename\n");
		    
		    # Now load the file from disk
		    console("Loading $tempdir/$filename\n");
		    open(INPUT, $tempdir . "/" . $filename);
		    $post_image = readline(\*INPUT);
		    $post_image_filename = $filename;
		    close(INPUT);
		    `/bin/rm $tempdir/"$filename"`;
		    
		    if (($filename =~ /jpg/i)
			|| ($filename =~ /gif/i)
			|| ($filename =~ /png/i)) {
			
			console("Found image name: $filename\n");
			db_execute("INSERT INTO $posts_images " .
				   "(post_id, post_image, " .
				   "filename, post_image_index) " .
				   "VALUES " .
				   "($postid, " .
				   $dbh->quote($post_image) .
				   ", '$post_image_filename', " .
				   "'$found_image')");
			$found_image++;
		    } elsif ($filename =~ /txt/i) {
			$body = $dbh->quote($post_image);
		    }
		}
		closedir(DIR);
		
		db_execute("INSERT INTO $posts_text " .
			   "(post_id, post_text) " .
			   "VALUES " .
			   "($postid, $body)");
		
		undef $found_image;
		`/bin/rmdir $tempdir`;
	    } else {
		db_execute("INSERT INTO $posts_text " .
			   "(post_id, post_text) " .
			   "VALUES " .
			   "($postid, $body)");
	    }
	    
	    
	    db_execute("UPDATE $forums " .
		       "SET forum_posts = forum_posts + 1, " . 
		       "forum_topics = forum_topics + 1 " . 
		       "WHERE forum_id = $forumid");
	    
	    db_execute("UPDATE $users " .
		       "SET user_posts = user_posts + 1 " .
		       "WHERE user_id = $newsgatewayid");
	}
	db_execute("DELETE FROM phpbb_usenet_article WHERE msgid = $msgid");
	db_execute("DELETE FROM phpbb_usenet_ref WHERE msgid = $msgid");
    }
    db_execute("commit");
    $q1->finish;
    console(" -> inserted $tcount threads\n");
}

sub console {
    my ($output) = @_;

    if ($debug) {
	print "$output";
    }
}

sub db_connect {
    my $dsn = "DBI:mysql:database=$db;host=$db_host";
    my $dbh = DBI->connect( $dsn,
			    $db_username,
			    $db_password,
			    { RaiseError => 1,
			      PrintError => 0}
			    ) 
	|| die "Cant connect to $db using $db_username\@$db_host";
    return $dbh;
}

sub db_fetch {
    my $sql = $_[0];
    my $qry;
	
    if ($debug > 1) {
	console($sql . "\n");
    }

    $qry = $dbh->prepare($sql) or die $!;
    $qry->execute() or die "Query failed ($sql) $!";
    return $qry;
}

sub db_execute {
    my $sql = $_[0];
    my $ignore_when_fail = $_[1];
	
    if ($debug > 1) {
	console($sql . "\n");
    }
    eval { $dbh->do($sql) };
    if ($@) {
	if (defined($ignore_when_fail)) { 
	    return 0;
	} else {
	    die "\nQuery failed:\n$sql\n\n$@";
	}
    }
    return 1;
}

sub article_batch_process {
    my $a = $_[0];
    my $forum_id = $_[1];
    my $forum_name = $_[2];
    my ($lastmsg);
    my $cnt = 0;
    my $missing = 0;
    my @cleanarticles = ();
    my @ra = ();

    console(" -> processing article batch...\n");
    db_execute("begin");
    foreach my $art (@$a) {
	my $threadid = 0;
	my $postid =0;
	$cnt++;
	my $postexists = 0;
	my $r = $$art{refs};
	my $b = $$art{body};
	if ($b) {
	    $$art{body} = join("",@$b);
	    # $$art{body} =~ s/begin\s+666.*\nend\n//gsoi;
	    $$art{body} = $wd->decode($$art{body});
	    $$art{body} = article_cleanup($$art{body});
	}
	if ($$art{refs}) {
	    $$art{refs} =~ s/,|\s//go; 
	    $$art{refs} =~ s/></> </go;
	} else {
	    if ($$art{subject}) {
		$$art{subject} =~ s/\</\(/go; 
		$$art{subject} =~ s/\>/\)/go;
		$$art{subject} =~ s/^\s+//gso; 
		$$art{subject} =~ s/\s+$//gso;
		if ($$art{subject} =~ /^Re:\s+(.*)/o) {
		    my $qry = db_fetch("SELECT msgid " .
				       "FROM $topics " .
				       "WHERE forum_id = $forum_id " .
				       "AND topic_title = " . $dbh->quote($1) . " " .
				       "ORDER BY topic_time " .
				       "LIMIT 1");
		    my $msgid = $qry->fetchrow_array; 
		    $qry->finish;
		    if($msgid) {
			$$art{refs} = "$msgid";
		    }
		}
	    }
	}
	$lastmsg = $$art{msgnum};
	my $forumid    = $forum_id;
	my $msgnum     = $$art{msgnum};
	if (!$$art{subject} || $$art{subject} =~ /^\s+$/o) {
	    $$art{subject} = "[No Subject]";
	}
	if ($$art{msgid} && $$art{body} && $$art{msgnum}) {
	    my $group      = $dbh->quote($forum_name);
	    my $msgid      = $dbh->quote($$art{msgid});
	    my $nntpdtm    = $$art{nntpdtm};
	    if ($nntpdtm > time()+600) {
		$nntpdtm = time();
	    }
	    $nntpdtm       = $dbh->quote($nntpdtm);
	    my $subject    = $dbh->quote($$art{subject});
	    my $poster     = $dbh->quote($$art{poster});
	    my $email      = $dbh->quote($$art{email});
	    my $refs       = $dbh->quote($$art{refs});
	    my $body       = $dbh->quote($$art{body});
	    my $nntpposter = $dbh->quote(encode_entities($$art{from}));
	    if ($$art{refs}) {
		@ra = split(/\s/,$$art{refs});
		my $tmp = "";
		foreach my $ref (@ra) {
		    my $re = $dbh->quote($ref);
		    $tmp .= "($msgid,$re,$#ra+1,$nntpdtm),";
		}
		chop($tmp);
		db_execute("INSERT IGNORE INTO phpbb_usenet_ref " .
			   "(msgid,ref,cnt,dtm) " .
			   "VALUES $tmp",1);
	    }
	    if ($$art{useragent} && ($$art{useragent} =~ /$useragent/)) {
		($threadid,$postid) = split(/\|/,$$art{xref});
		my $dbpostexists = db_fetch("SELECT count(*) " .
					    "FROM $posts WHERE " .
					    "post_id = '$postid'");
		$postexists = $dbpostexists->fetchrow_array; 
		$dbpostexists->finish;
		if ($postexists) {
		    db_execute("UPDATE $posts " .
			       "SET msgid = $msgid " .
			       "WHERE post_id = $postid");
		    db_execute("UPDATE phpbb_usenet_outgoing_log " .
			       "SET msgid = $msgid " .
			       "WHERE postid = $postid");
		    db_execute("UPDATE phpbb_usenet_outgoing " .
			       "SET refs = $msgid " .
			       "WHERE threadid = $threadid");
		    db_execute("UPDATE phpbb_usenet_outgoing " .
			       "SET refs = concat(refs,' ',$msgid) " .
			       "WHERE inreplyto = $postid");
		    
		    if (!$$art{refs}) {
			db_execute("UPDATE $topics " .
				   "SET msgid = $msgid " .
				   "WHERE topic_id = $threadid");
			db_execute("UPDATE phpbb_usenet_outgoing " .
				   "SET refs = $msgid " .
				   "WHERE threadid = $threadid");
			db_execute("UPDATE phpbb_usenet_outgoing " .
				   "SET refs = concat(refs,' ',$msgid) " .
				   "WHERE inreplyto = $postid");
		    }
		    db_execute("DELETE FROM phpbb_usenet_article " .
			       "WHERE msgid = $msgid");
		    db_execute("DELETE FROM phpbb_usenet_ref " .
			       "WHERE msgid = $msgid");
		}
	    }
	    unless ($postexists || $postid) {
		db_execute("INSERT IGNORE INTO phpbb_usenet_article " .
			   "(newsgroup,forum,msgid,dtm,subject,poster," .
			   "email,refs,body,msgnum,nntpposter," .
			   "ord_key,threadid,postid) " .
			   "VALUES " .
			   "($group,$forumid,$msgid,$nntpdtm,$subject," .
			   "$poster,$email,$refs,$body,$msgnum,$nntpposter," .
			   "$#ra+1,$threadid,$postid)", 1);
	    }
	} else {
	    $missing++;
	}
    }
    db_execute("commit");
}

sub article_expire {
   my $expire = time() - ($expire_days * 86400);
   db_execute("begin");
   db_execute("DELETE FROM phpbb_usenet_article WHERE dtm < $expire");
   db_execute("DELETE FROM phpbb_usenet_ref WHERE dtm < $expire");
   db_execute("commit");
}


sub article_quote {
   my $fbody = $_[0];
   my @ibody = ();
   my @abody = split(/\n/,$fbody);
   foreach my $albody (@abody) {
#      if ($albody =~ /^>>>|^> > >/o) {
      if ($albody =~ /^&gt;&gt;&gt;|^&gt; &gt; &gt;/o) {
         push(@ibody,'<font size="1" color="' .
			  $quote3 . '">' . $albody . "</font><br />");
#	  $albody =~ s/&gt;//gio;
#	  push(@ibody, '[quote:0=&quot;Previous poster&quot;]' . $albody . "[/quote]\n");
#      } elsif ($albody =~ /^>>|^> >/o) {
      } elsif ($albody =~ /^&gt; &gt;|^&gt; &gt;/o) {
         push(@ibody,'<font size="1" color="' .
			  $quote2 . '">' . $albody . "</font><br />");
#	  $albody =~ s/&gt;//gio;
#	  push(@ibody, '[quote:0=&quot;Previous poster&quot;]' . $albody . "[/quote]\n");
#      } elsif ($albody =~ /^>|^:|^\|/o) {
      } elsif ($albody =~ /^&gt|^:|^\|/o) {
         push(@ibody,'<font size="1" color="' . 
			  $quote1 . '">' . $albody . "</font><br />");
#	  $albody =~ s/&gt;//gio;
#	  push(@ibody,'[quote:0=&quot;Previous poster&quot]' . $albody . "[/quote]\n");
      } else {
         push(@ibody,$albody."<br>");
      }
   }
   $fbody = join("",@ibody);
   # Remove lines that quote & unquote over again
   #$fbody =~ s#\[/quote\]\n\[quote:0=&quot;Previous poster&quot;\]##gios;
   #$fbody =~ s#\[/quote\]\n\[quote:0="Previous poster"\]##gios;
   return $fbody;
}

sub article_add_html {
   my $text = $_[0];
   $text =~ s/^\s+//gio; 
   $text =~ s/\s+$//gio; 
   $text =~ s/\n+$//gio;
   $text =~ s/>$//gos; 
   $text =~ s/^\s+//gio; 
   $text =~ s/\s+$//gio;
   $text =~ s/\n+$//gio;
   $text =~ s/\n\n\n/\n\n/gso; 
   $text =~ s/\n\n\n/\n\n/gso;

   # Make all HTML just "show up"
   $text =~ s/</&lt;/gso;
   $text =~ s/>/&gt;/gso;

   if ($hyperlinks) {
       find_uris($text, sub { my ($uri, $orig_uri) = @_;
			      return '<A HREF="'.$uri.'">'.$orig_uri.'</A>';
			  });
   }
   if ($custom_quotes) {
      $text = article_quote($text);
   }

   return $text;
}

sub article_cleanup {
    my $s = $_[0];

    $s =~ s/^\s+//gso; 
    $s =~ s/\s+$//gso; 
    $s =~ s/^://so; 
    $s =~ s/^\s+//go;
    $s =~ s/\s+$//go; 
    $s =~ s/^\n$//go; 
    $s =~ s/^>$//gso; 
    $s =~ s/^:$//go;
    $s =~ s/^\:$//go; 
    $s =~ s/\n\n\n/\n\n/go; 
    $s =~ s/>$//gso;
    $s =~ s/\s+>$//gso; 
    $s =~ s/^>\s+\n//gso; 
    $s =~ s/^>\n//gso;
    $s =~ s/^\s+>\s+\n//gso; 
    $s =~ s/^\s+>\n//gso;

    #$s =~ s/</&lt;/gso;
    #$s =~ s/>/&gt;/gso;
    return $s;
}

sub connect_usenet {
    my $newsgroup = $_[0];
    my ($server, $port);

    if ($$newsgroup->{server} =~ /:/) {
	my $connectstring = $$newsgroup->{server}; 
	$connectstring =~ /(.*):(.*)/;
	$server = $1; 
	$port = $2;
    } else {
	$server = $$newsgroup->{server}; 
	$port = '119';
    }
    console("Connecting to $$newsgroup->{server}... \n");
    my $con = new News::NNTPClient($server,$port) 
	|| die "Unable to connect to $$newsgroup->{server}";
    console("Connected\n");
    if ($$newsgroup->{username} && $$newsgroup->{password}) {
	console("Sending authentication info... \n");
	$con->authinfo("$$newsgroup->{username}","$$newsgroup->{password}") 
	    || die "Authentication failure";
	console("Authenticated and logged in\n");
    }
    return $con;
}

sub remove_bb_code {
    my $text = $_[0];
    my ($bbo,$bbc);
    $text =~ s/&amp;/&/ig;                                                # ampersand
    $text =~ s/&quot;/\"/ig;
    $text =~ s/&lt;/</ig;
    $text =~ s/&gt;/>/ig;
    $text =~ s/\[size=[0-9]+\]|\[\/size\]//ig;                            # size
    $text =~ s/\[color=(\"\#)?[A-Za-z0-9]+(\")?\]|\[\/color\]//ig;        # color
    $text =~ s/\[url(=)?(\")?//ig;
    $text =~ s/(\")?\](.+)\[\/url\]/$2/gi;
    $text =~ s/\[email(=)?(\")?//ig;
    $text =~ s/(\")?\](.+)\[\/email\]/$2/gi;
    $text =~ s/\[font=(\"\#)?[A-Za-z]+(\")?\]|\[\/font\]//ig;             # font
    $text =~ s/\[list(=)?[1Aa]?\]|\[\/list(=)?[1Aa]?\]//ig;               # list
    $text =~ s/\[\*\]/ - /ig;
    $text =~ s/\[(\/)?code\]//ig;
    
    $text =~ s/\r/\n/g;

    # strip out html linefeeds and replace them with real ones
    $text =~ s/<BR>/\n/gi;
    $text =~ s/<BR \/>/\n/gi;
    
    # strip out all other html
    $text =~ s/<.*?>//gi;

    return $text;
}


sub get_batch {
    my @m = ($_[0],$_[1]);
    my $c = $_[2];
    my @a=();
    my $h;

    console(" -> fetching articles $m[0] to $m[1]...\n");
    my $mid = $c->stat($m[0]);
    for (my $msg=$m[0]+1;$msg<=$m[1];$msg++) {
	undef($h);
	my %article=();

	if ($h = $c->head($mid)) {
	    console(" -> fetching $msg... ");
	    my($f,$s,$d,$i,$r,$u,$xr,$xn,$l,$p,$e,$xa,$b);
	    foreach (@$h) {
		if (/^From: (.*)/)         { $f  = $wd->decode($1); }
		if (/^Subject: (.*)/)      { $s  = $wd->decode($1); }
		if (/^Date: (.*)/)         { $d  = str2time($1); }
		if (/^Message-ID: (.*)/)   { $i  = $1; }
		if (/^References: (.*)/)   { $r  = $1; }
		if (/^User-Agent: (.*)/)   { $u  = $1; }
		if (/^X-References: (.*)/) { $xr = $1; }
		if (/^X-Newsreader: (.*)/) { $xn = $1; }
		if (/^Lines: (.*)/)        { $l  = $1; }
		if (/^X-No-Archive: (.*)/) { $xa = $1; }
		if (defined($f)) {
		    decode_entities($f);
		}
		if (defined($s)) {
		    decode_entities($s);
		}
	    }
	    if (!$d || $d=~/^\s+/) {
		$d=time();
	    }
	    ($p,$e) = parse_from($f);
	    my $spam = 0;
	    if (($l && $l < $maxmsglen) || (!$l)) {
		if (!$spam) {
		    undef $b;
		    if ($b = $c->body($mid)) {
			push (@a,{ from      => $f,
				   subject   => $s,
				   nntpdtm   => $d,
				   msgid     => $i,
				   refs      => $r,
				   useragent => $u,
				   xref      => $xr,
				   xreader   => $xn,
				   lines     => $l,
				   poster    => $p,
				   email     => $e,
				   msgnum    => $msg,
				   body      => $b
				   }
			      );
			
			if ($debug >= 2) {
			    console(" OK - $f - $s\n");
			} else {
			    console(" OK \n");
			}
		    }
		} else {
		    console(" Not fetched - Spam\n");
		}
	    } else {
		console(" Not fetched - Exceeds maxmsglen\n");
	    }
	}
	$mid = $c->next();
	last if (!$mid);
    }
    return @a;
}

sub parse_from {
    my $from = $_[0];
    my ($email,$name,$qe,$epref);
    if (defined($from)) {
	$from =~ s/^\s+//gos; $from =~ s/\s+$//gos;
	$from =~ s/\[/-lsq/g; $from =~ s/\[/-rsq/g;
	if (my @mail_address = Mail::Address->parse($from)) {
	    $email = $mail_address[0]->address();
	    $from = $mail_address[0]->name();
	    if (!$from){
		$from = $mail_address[0]->user();
	    }
	    $from =~ s/-rsq/[/g; $from =~ s/-lsq/]/g;
	    $from = substr($from,0,$maxuserlength);
	}
    } else {
	$from = "Anonymous";
    }
    return ($from,$email);
}


1;
